/*
 * MIT License
 * Copyright (C) 2022 4Auth Limited. All rights reserved
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.bureau.onetaplogin

import android.os.Build
import org.json.JSONException
import org.json.JSONObject
import java.io.BufferedReader
import java.io.InputStreamReader
import java.io.OutputStream
import java.net.HttpCookie
import java.net.Socket
import java.net.URL
import java.nio.charset.Charset
import java.nio.charset.StandardCharsets
import java.util.UUID
import javax.net.ssl.SSLSocketFactory

internal class ClientSocket {
    private lateinit var socket: Socket
    private lateinit var output: OutputStream
    private lateinit var input: BufferedReader

    fun open(url: URL, accessToken: String?, operator: String?): JSONObject {
        val requestId: String = UUID.randomUUID().toString()
        var redirectURL: URL? = null
        var redirectCount = 0
        var result: ResultHandler? = null
        do {
            redirectCount += 1
            val nurl = redirectURL ?: url
            try {
                startConnection(nurl)
                result = if (redirectCount == 1) {
                    sendCommand(nurl, accessToken, operator, null, requestId)
                } else {
                    sendCommand(nurl, null, null, result?.getCookies(), requestId)
                }
                redirectURL = if (result?.getRedirect() != null) {
                    result.getRedirect()
                } else {
                    null
                }
                stopConnection()
            } catch (ex: Exception) {
                return convertError("sdk_connection_error", "ex: ".plus(ex.localizedMessage))
            }
        } while (redirectURL != null && redirectCount <= MAX_REDIRECT_COUNT)
        if (redirectCount == MAX_REDIRECT_COUNT) {
            return convertError("sdk_redirect_error", "Too many redirects")
        }
        if (result != null) {
            return convertResultHandler(result)
        }
        return convertError("sdk_error", "internal error")
    }

    private fun convertResultHandler(res: ResultResponse): JSONObject {
        val json = JSONObject()
        json.put("http_status", res.getHttpStatus())
        try {
            if (res.getBody() != null) {
                json.put("response_body", JSONObject(res.getBody()))
            }
            return json
        } catch (e: JSONException) {
            if (res.getBody() != null) {
                json.put("response_raw_body", res.getBody())
            } else {
                return convertError("sdk_error", "ex: ".plus(e.localizedMessage))
            }
        }
        return json
    }

    private fun convertError(code: String, description: String): JSONObject {
        val json = JSONObject()
        json.put("error", code)
        json.put("error_description", description)
        return json
    }

    private fun makePost(url: URL, headers: Map<String, String>, body: String?): String {
        val cmd = StringBuffer()
        cmd.append("POST " + url.path)
        cmd.append(" HTTP/1.1$CRLF")
        cmd.append("Host: " + url.host)
        if (url.protocol == "https" && url.port > 0 && url.port != PORT_443) {
            cmd.append(":" + url.port)
        } else if (url.protocol == "http" && url.port > 0 && url.port != PORT_80) {
            cmd.append(":" + url.port)
        }
        cmd.append(CRLF)
        headers.forEach { entry ->
            cmd.append(entry.key + ": " + entry.value + CRLF)
        }
        if (body != null) {
            cmd.append("Content-Length: " + body.length + CRLF)
            cmd.append("Connection: close$CRLF$CRLF")
            cmd.append(body)
            cmd.append("$CRLF$CRLF")
        } else {
            cmd.append("Content-Length: 0$CRLF")
            cmd.append("Connection: close$CRLF$CRLF")
        }
        return cmd.toString()
    }

    private fun sendAndReceive(request: String): ResponseHandler? {
        try {
            val bytesOfRequest: ByteArray =
                request.toByteArray(Charset.forName(StandardCharsets.UTF_8.name()))
            output.write(bytesOfRequest)
            output.flush()
        } catch (ex: Exception) {
            throw ex
        }
        var status = 0
        var body = String()
        var result: ResponseHandler? = null
        var chunked = false
        try {
            val response: String? = input.use { it.readText() }
            response?.let {
                val lines = response.split("\n")
                for (line in lines) {
                    if (line.startsWith("HTTP/")) {
                        val parts = line.split(" ")
                        if (parts.isNotEmpty() && parts.size >= 2) {
                            status = Integer.valueOf(parts[1].trim())
                        }
                    } else if (line.startsWith("Transfer-Encoding:")) {
                        val parts = line.split(" ")
                        if (parts.isNotEmpty() && parts.size > 1) {
                            if (parts[1].contains("chunked")) chunked = true
                        }
                    } else if (line.contains(": ") && body.isEmpty()) {
                        // do nothing
                    } else {
                        body += line.replace("\r", "")
                    }
                }
                if (chunked && body.isNotBlank()) {
                    val r1: Int = body.indexOf("{")
                    val r2: Int = body.lastIndexOf("}")
                    if (r1 in 1 until r2) {
                        body = body.substring(r1, r2 + 1)
                    }
                }
                result = ResponseHandler(status, body)
            }
        } catch (ex: Exception) {
            throw ex
        }
        return result
    }

    fun post(url: URL, headers: Map<String, String>, body: String?): JSONObject {
        startConnection(url)
        val request = makePost(url, headers, body)
        val response = sendAndReceive(request)
        stopConnection()
        if (response != null) {
            return convertResultHandler(response)
        }
        return convertError("sdk_error", "internal error")
    }

    private fun makeHTTPCommand(url: URL, accessToken: String?, operator: String?, cookies: ArrayList<HttpCookie>?, requestId: String?): String {
        val cmd = StringBuffer()
        cmd.append("GET " + url.path)
        if (url.path.isEmpty()) {
            cmd.append("/")
        }
        if (url.query != null) {
            cmd.append("?" + url.query)
        }
        cmd.append(" HTTP/1.1$CRLF")
        cmd.append("Host: " + url.host)
        if (url.protocol == "https" && url.port > 0 && url.port != PORT_443) {
            cmd.append(":" + url.port)
        } else if (url.protocol == "http" && url.port > 0 && url.port != PORT_80) {
            cmd.append(":" + url.port)
        }
        cmd.append(CRLF)
        if (accessToken != null) {
            cmd.append("Authorization: Bearer ${accessToken}$CRLF")
        }
        if (requestId != null) {
            cmd.append("x-tru-sdk-request: ${requestId}$CRLF")
        }
        if (operator != null) {
            cmd.append("x-tru-ops: ${operator}$CRLF")
        }
        if (isEmulator()) {
            cmd.append("x-tru-mode: sandbox$CRLF")
        }

        cmd.append("Accept: text/html,application/xhtml+xml,application/xml,*/*$CRLF")
        val cs = StringBuffer()
        var cookieCount = 0
        val iterator = cookies.orEmpty().listIterator()
        for (cookie in iterator) {
            if (((cookie.secure && url.protocol == "https") || (!cookie.secure)) &&
                (cookie.domain == null || (cookie.domain != null && url.host.contains(cookie.domain))) &&
                (cookie.path == null || url.path.startsWith(cookie.path))
            ) {
                if (cookieCount > 0) cs.append("; ")
                cs.append(cookie.name + "=" + cookie.value)
                cookieCount++
            }
        }
        if (cs.length > 1) cmd.append("Cookie: $cs$CRLF")

        cmd.append("Connection: close$CRLF$CRLF")
        return cmd.toString()
    }

    private fun sendCommand(url: URL, accessToken: String?, operator: String?, cookies: ArrayList<HttpCookie>?, requestId: String?): ResultHandler? {
        val command = makeHTTPCommand(url, accessToken, operator, cookies, requestId)
        return sendAndReceive(url, command, cookies)
    }

    private fun startConnection(url: URL) {
        var port = PORT_80
        if (url.port > 0) port = url.port
        try {
            socket = if (url.protocol == "https") {
                port = PORT_443
                if (url.port > 0) port = url.port
                SSLSocketFactory.getDefault().createSocket(url.host, port)
            } else {
                Socket(url.host, port)
            }
        } catch (ex: Exception) {
            throw ex
        }
        return try {
            socket.soTimeout = 5 * 1000
            output = socket.getOutputStream()
            input = BufferedReader(InputStreamReader(socket.inputStream))
        } catch (ex: Exception) {
            if (!socket.isClosed) socket.close()
            throw ex
        }
    }

    private fun sendAndReceive(requestURL: URL, message: String, existingCookies: ArrayList<HttpCookie>?): ResultHandler? {
        try {
            val bytesOfRequest: ByteArray =
                message.toByteArray(Charset.forName(StandardCharsets.UTF_8.name()))
            output.write(bytesOfRequest)
            output.flush()
        } catch (ex: Exception) {
            throw ex
        }
        var status = 0
        var body: String? = null
        var type = ""
        var result: ResultHandler? = null
        var bodyBegin = false
        val cookies: ArrayList<HttpCookie> = ArrayList()
        if (existingCookies != null) cookies.addAll(existingCookies)

        try {
            // convert the entire stream in a String
            val response: String? = input.use { it.readText() }
            response?.let {
                val lines = response.split("\n")
                for (line in lines) {
                    if (line.startsWith("HTTP/")) {
                        val parts = line.split(" ")
                        if (parts.isNotEmpty() && parts.size >= 2) {
                            status = Integer.valueOf(parts[1].trim())
                        }
                    } else if (line.contains("Set-Cookie:") || line.contains("set-cookie:")) {
                        val parts: List<String> = line.split("ookie:")
                        if (parts.isNotEmpty() && parts.size > 1) {
                            try {
                                for (cookie in HttpCookie.parse(parts[1])) {
                                    cookies.add(cookie)
                                }
                            } catch (ex: IllegalArgumentException) {
                            }
                        }
                    } else if (line.contains("Location:") || line.contains("location:")) {
                        result = parseRedirect(status, requestURL, line.replace("\r", ""), cookies)
                    } else if (line.contains("Content-Type:")) {
                        val parts = line.split(" ")
                        if (parts.isNotEmpty() && parts.size > 1) {
                            type = parts[1].replace(";", "").replace("\r", "")
                        }
                    } else if (("application/json" == type || "application/hal+json" == type || "application/problem+json" == type) && line == "\r") {
                        bodyBegin = true
                    } else if (bodyBegin) {
                        body = if (body != null) body + line.replace("\r", "") else line.replace("\r", "")
                    }
                }
                result = if (result != null) {
                    return result
                } else if (bodyBegin && body != null && body != "null") {
                    ResultHandler(status, null, parseBodyIntoJSONString(body), cookies)
                } else {
                    ResultHandler(status, null, null, null)
                }
            }
        } catch (ex: Exception) {
            throw ex
        }
        return result
    }

    private fun parseBodyIntoJSONString(body: String?): String? {
        if (body != null) {
            val start = body.indexOf("{")
            val end = body.lastIndexOf("}")
            return body.subSequence(start, end + 1).toString()
        }
        return null
    }

    private fun parseRedirect(httpStatus: Int, requestURL: URL, redirectLine: String, cookies: ArrayList<HttpCookie>?): ResultHandler? {
        val parts = redirectLine.split("ocation: ")
        if (parts.isNotEmpty() && parts.size > 1) {
            if (parts[1].isBlank()) return null
            val redirect = parts[1]
            // some location header are not properly encoded
            val cleanRedirect = redirect.replace(" ", "+")
            if (!cleanRedirect.startsWith("http")) { // http & https
                return ResultHandler(httpStatus, URL(requestURL, cleanRedirect), null, cookies)
            }
            return ResultHandler(httpStatus, URL(cleanRedirect), null, cookies)
        }
        return null
    }

    private fun stopConnection() {
        try {
            input.close()
            output.close()
            socket.close()
        } catch (e: Throwable) {
        }
    }

    private fun isEmulator(): Boolean {
        return Build.FINGERPRINT.contains("generic") ||
            Build.FINGERPRINT.startsWith("unknown") ||
            Build.MODEL.contains("google_sdk") ||
            Build.MODEL.contains("Emulator") ||
            Build.MODEL.contains("Android SDK built for x86") ||
            Build.MANUFACTURER.contains("Genymotion") ||
            (Build.BRAND.startsWith("generic") && Build.DEVICE.startsWith("generic")) ||
            Build.PRODUCT.contains("sdk_gphone_x86")
    }

    companion object {
        private const val TAG = "CellularClient"
        private const val HEADER_USER_AGENT = "User-Agent"
        private const val MAX_REDIRECT_COUNT = 10
        private const val PORT_80 = 80
        private const val PORT_443 = 443
        private const val CRLF = "\r\n"
    }

    class ResultHandler(httpStatus: Int, redirect: URL?, body: String?, cookies: ArrayList<HttpCookie>?) : ResultResponse {
        private val s: Int = httpStatus
        private val r: URL? = redirect
        private val b: String? = body
        private val cs: ArrayList<HttpCookie>? = cookies

        override fun getHttpStatus(): Int {
            return s
        }

        fun getRedirect(): URL? {
            return r
        }

        override fun getBody(): String? {
            return b
        }

        fun getCookies(): ArrayList<HttpCookie>? {
            return cs
        }
    }

    class ResponseHandler(httpStatus: Int, body: String?) : ResultResponse {
        private val s: Int = httpStatus
        private val b: String? = body

        override fun getHttpStatus(): Int {
            return s
        }

        override fun getBody(): String? {
            return b
        }
    }

    internal interface ResultResponse {
        fun getHttpStatus(): Int

        fun getBody(): String?
    }
}
